"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const events_1 = require("events");
const test_utils_1 = require("./test-utils");
const client_1 = __importDefault(require("./client"));
const errors_1 = require("./errors");
const lua_script_1 = require("./lua-script");
const sinon_1 = require("sinon");
describe('Client', () => {
    describe('authentication', () => {
        test_utils_1.itWithClient(test_utils_1.TestRedisServers.PASSWORD, 'Client should be authenticated', async (client) => {
            assert_1.strict.equal(await client.ping(), 'PONG');
        });
        it('should not retry connecting if failed due to wrong auth', async () => {
            const client = client_1.default.create({
                socket: {
                    ...test_utils_1.TEST_REDIS_SERVERS[test_utils_1.TestRedisServers.PASSWORD],
                    password: 'wrongpassword'
                }
            });
            await assert_1.strict.rejects(client.connect(), {
                message: 'WRONGPASS invalid username-password pair or user is disabled.'
            });
            assert_1.strict.equal(client.isOpen, false);
        });
    });
    describe('legacyMode', () => {
        const client = client_1.default.create({
            socket: test_utils_1.TEST_REDIS_SERVERS[test_utils_1.TestRedisServers.OPEN],
            legacyMode: true
        });
        before(() => client.connect());
        afterEach(() => client.modern.flushAll());
        after(() => client.disconnect());
        it('client.sendCommand should call the callback', done => {
            client.sendCommand('PING', (err, reply) => {
                if (err) {
                    return done(err);
                }
                try {
                    assert_1.strict.equal(reply, 'PONG');
                    done();
                }
                catch (err) {
                    done(err);
                }
            });
        });
        it('client.sendCommand should work without callback', async () => {
            client.sendCommand('PING');
            await client.modern.ping(); // make sure the first command was replied
        });
        it('client.modern.sendCommand should return a promise', async () => {
            assert_1.strict.equal(await client.modern.sendCommand(['PING']), 'PONG');
        });
        it('client.{command} should accept vardict arguments', done => {
            client.set('a', 'b', (err, reply) => {
                if (err) {
                    return done(err);
                }
                try {
                    assert_1.strict.equal(reply, 'OK');
                    done();
                }
                catch (err) {
                    done(err);
                }
            });
        });
        it('client.{command} should accept arguments array', done => {
            client.set(['a', 'b'], (err, reply) => {
                if (err) {
                    return done(err);
                }
                try {
                    assert_1.strict.equal(reply, 'OK');
                    done();
                }
                catch (err) {
                    done(err);
                }
            });
        });
        it('client.{command} should accept mix of strings and array of strings', done => {
            client.set(['a'], 'b', ['GET'], (err, reply) => {
                if (err) {
                    return done(err);
                }
                try {
                    assert_1.strict.equal(reply, null);
                    done();
                }
                catch (err) {
                    done(err);
                }
            });
        });
        it('client.multi.exec should call the callback', done => {
            client.multi()
                .ping()
                .exec((err, reply) => {
                if (err) {
                    return done(err);
                }
                try {
                    assert_1.strict.deepEqual(reply, ['PONG']);
                    done();
                }
                catch (err) {
                    done(err);
                }
            });
        });
        it('client.multi.exec should work without callback', async () => {
            client.multi()
                .ping()
                .exec();
            await client.modern.ping(); // make sure the first command was replied
        });
        it('client.modern.exec should return a promise', async () => {
            assert_1.strict.deepEqual(await (client.multi().modern
                .ping()
                .exec()), ['PONG']);
        });
    });
    describe('events', () => {
        it('connect, ready, end', async () => {
            const client = client_1.default.create({
                socket: test_utils_1.TEST_REDIS_SERVERS[test_utils_1.TestRedisServers.OPEN]
            });
            await Promise.all([
                client.connect(),
                events_1.once(client, 'connect'),
                events_1.once(client, 'ready')
            ]);
            await Promise.all([
                client.disconnect(),
                events_1.once(client, 'end')
            ]);
        });
    });
    describe('sendCommand', () => {
        test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'PING', async (client) => {
            assert_1.strict.equal(await client.sendCommand(['PING']), 'PONG');
        });
        describe('AbortController', () => {
            test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'success', async (client) => {
                await client.sendCommand(['PING'], {
                    signal: new AbortController().signal
                });
            });
            test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'AbortError', async (client) => {
                const controller = new AbortController();
                controller.abort();
                await assert_1.strict.rejects(client.sendCommand(['PING'], {
                    signal: controller.signal
                }), errors_1.AbortError);
            });
        });
    });
    describe('multi', () => {
        test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'simple', async (client) => {
            assert_1.strict.deepEqual(await client.multi()
                .ping()
                .set('key', 'value')
                .get('key')
                .exec(), ['PONG', 'OK', 'value']);
        });
        test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'should reject the whole chain on error', async (client) => {
            client.on('error', () => {
                // ignore errors
            });
            await assert_1.strict.rejects(client.multi()
                .ping()
                .addCommand(['DEBUG', 'RESTART'])
                .ping()
                .exec());
        });
        it('with script', async () => {
            const client = client_1.default.create({
                scripts: {
                    add: lua_script_1.defineScript({
                        NUMBER_OF_KEYS: 0,
                        SCRIPT: 'return ARGV[1] + 1;',
                        transformArguments(number) {
                            assert_1.strict.equal(number, 1);
                            return [number.toString()];
                        },
                        transformReply(reply) {
                            assert_1.strict.equal(reply, 2);
                            return reply;
                        }
                    })
                }
            });
            await client.connect();
            try {
                assert_1.strict.deepEqual(await client.multi()
                    .add(1)
                    .exec(), [2]);
            }
            finally {
                await client.disconnect();
            }
        });
    });
    it('scripts', async () => {
        const client = client_1.default.create({
            scripts: {
                add: lua_script_1.defineScript({
                    NUMBER_OF_KEYS: 0,
                    SCRIPT: 'return ARGV[1] + 1;',
                    transformArguments(number) {
                        assert_1.strict.equal(number, 1);
                        return [number.toString()];
                    },
                    transformReply(reply) {
                        assert_1.strict.equal(reply, 2);
                        return reply;
                    }
                })
            }
        });
        await client.connect();
        try {
            assert_1.strict.equal(await client.add(1), 2);
        }
        finally {
            await client.disconnect();
        }
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'should reconnect after DEBUG RESTART', async (client) => {
        client.on('error', () => {
            // ignore errors
        });
        await client.sendCommand(['CLIENT', 'SETNAME', 'client']);
        await assert_1.strict.rejects(client.sendCommand(['DEBUG', 'RESTART']));
        assert_1.strict.ok(await client.sendCommand(['CLIENT', 'GETNAME']) === null);
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'should SELECT db after reconnection', async (client) => {
        client.on('error', () => {
            // ignore errors
        });
        await client.select(1);
        await assert_1.strict.rejects(client.sendCommand(['DEBUG', 'RESTART']));
        assert_1.strict.equal((await client.clientInfo()).db, 1);
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'scanIterator', async (client) => {
        const promises = [], keys = [];
        for (let i = 0; i < 100; i++) {
            const key = i.toString();
            keys.push(key);
            promises.push(client.set(key, ''));
        }
        await Promise.all(promises);
        const set = new Set();
        for await (const key of client.scanIterator()) {
            set.add(key);
        }
        assert_1.strict.deepEqual([...set].sort(), keys.sort());
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'sScanIterator', async (client) => {
        const keys = [];
        for (let i = 0; i < 100; i++) {
            keys.push(i.toString());
        }
        await client.sAdd('key', keys);
        const set = new Set();
        for await (const key of client.sScanIterator('key')) {
            set.add(key);
        }
        assert_1.strict.deepEqual([...set].sort(), keys.sort());
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'zScanIterator', async (client) => {
        const members = [];
        for (let i = 0; i < 100; i++) {
            members.push({
                score: 1,
                value: i.toString()
            });
        }
        await client.zAdd('key', members);
        const map = new Map();
        for await (const member of client.zScanIterator('key')) {
            map.set(member.value, member.score);
        }
        function sort(a, b) {
            return Number(b[0]) - Number(a[0]);
        }
        assert_1.strict.deepEqual([...map.entries()].sort(sort), members.map(member => [member.value, member.score]).sort(sort));
    });
    test_utils_1.itWithClient(test_utils_1.TestRedisServers.OPEN, 'PubSub', async (publisher) => {
        const subscriber = publisher.duplicate();
        await subscriber.connect();
        try {
            const channelListener1 = sinon_1.spy(), channelListener2 = sinon_1.spy(), patternListener = sinon_1.spy();
            await Promise.all([
                subscriber.subscribe('channel', channelListener1),
                subscriber.subscribe('channel', channelListener2),
                subscriber.pSubscribe('channel*', patternListener)
            ]);
            await Promise.all([
                test_utils_1.waitTillBeenCalled(channelListener1),
                test_utils_1.waitTillBeenCalled(channelListener2),
                test_utils_1.waitTillBeenCalled(patternListener),
                publisher.publish('channel', 'message')
            ]);
            assert_1.strict.ok(channelListener1.calledOnceWithExactly('message', 'channel'));
            assert_1.strict.ok(channelListener2.calledOnceWithExactly('message', 'channel'));
            assert_1.strict.ok(patternListener.calledOnceWithExactly('message', 'channel'));
            await subscriber.unsubscribe('channel', channelListener1);
            await Promise.all([
                test_utils_1.waitTillBeenCalled(channelListener2),
                test_utils_1.waitTillBeenCalled(patternListener),
                publisher.publish('channel', 'message')
            ]);
            assert_1.strict.ok(channelListener1.calledOnce);
            assert_1.strict.ok(channelListener2.calledTwice);
            assert_1.strict.ok(channelListener2.secondCall.calledWithExactly('message', 'channel'));
            assert_1.strict.ok(patternListener.calledTwice);
            assert_1.strict.ok(patternListener.secondCall.calledWithExactly('message', 'channel'));
            await subscriber.unsubscribe('channel');
            await Promise.all([
                test_utils_1.waitTillBeenCalled(patternListener),
                publisher.publish('channel', 'message')
            ]);
            assert_1.strict.ok(channelListener1.calledOnce);
            assert_1.strict.ok(channelListener2.calledTwice);
            assert_1.strict.ok(patternListener.calledThrice);
            assert_1.strict.ok(patternListener.thirdCall.calledWithExactly('message', 'channel'));
            await subscriber.pUnsubscribe();
            await publisher.publish('channel', 'message');
            assert_1.strict.ok(channelListener1.calledOnce);
            assert_1.strict.ok(channelListener2.calledTwice);
            assert_1.strict.ok(patternListener.calledThrice);
        }
        finally {
            await subscriber.disconnect();
        }
    });
});
