"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cluster_key_slot_1 = __importDefault(require("cluster-key-slot"));
const client_1 = __importDefault(require("./client"));
class RedisClusterSlots {
    #options;
    #clientByKey = new Map();
    #slots = [];
    constructor(options) {
        this.#options = options;
    }
    async connect() {
        for (const rootNode of this.#options.rootNodes) {
            try {
                await this.#discoverNodes(rootNode);
                return;
            }
            catch (err) {
                // this.emit('error', err);
            }
        }
        throw new Error('None of the root nodes is available');
    }
    async discover(startWith) {
        try {
            await this.#discoverNodes(startWith.options?.socket);
            return;
        }
        catch (err) {
            // this.emit('error', err);
        }
        for (const client of this.#clientByKey.values()) {
            if (client === startWith)
                continue;
            try {
                await this.#discoverNodes(client.options?.socket);
                return;
            }
            catch (err) {
                // this.emit('error', err);
            }
        }
        throw new Error('None of the cluster nodes is available');
    }
    async #discoverNodes(socketOptions) {
        const client = client_1.default.create({
            socket: socketOptions
        });
        await client.connect();
        try {
            await this.#reset(await client.clusterNodes());
        }
        finally {
            await client.disconnect(); // TODO: catch error from disconnect?
        }
    }
    async #reset(masters) {
        // Override this.#slots and add not existing clients to this.#clientByKey
        const promises = [], clientsInUse = new Set();
        for (const master of masters) {
            const slot = {
                master: this.#initiateClientForNode(master, false, clientsInUse, promises),
                replicas: this.#options.useReplicas ?
                    master.replicas.map(replica => this.#initiateClientForNode(replica, true, clientsInUse, promises)) :
                    [],
                iterator: undefined // will be initiated in use
            };
            for (const { from, to } of master.slots) {
                for (let i = from; i < to; i++) {
                    this.#slots[i] = slot;
                }
            }
        }
        // Remove unused clients from this.#clientBykey using clientsInUse
        for (const [key, client] of this.#clientByKey.entries()) {
            if (clientsInUse.has(key))
                continue;
            // TODO: ignore error from `.disconnect`?
            promises.push(client.disconnect());
            this.#clientByKey.delete(key);
        }
    }
    #initiateClientForNode(node, readonly, clientsInUse, promises) {
        clientsInUse.add(node.url);
        let client = this.#clientByKey.get(node.url);
        if (!client) {
            client = client_1.default.create({
                socket: {
                    host: node.host,
                    port: node.port
                },
                readonly
            });
            promises.push(client.connect());
            this.#clientByKey.set(node.url, client);
        }
        return client;
    }
    #getSlotMaster(slot) {
        return this.#slots[slot].master;
    }
    *#slotIterator(slotNumber) {
        const slot = this.#slots[slotNumber];
        yield slot.master;
        for (const replica of slot.replicas) {
            yield replica;
        }
    }
    #getSlotClient(slotNumber) {
        const slot = this.#slots[slotNumber];
        if (!slot.iterator) {
            slot.iterator = this.#slotIterator(slotNumber);
        }
        const { done, value } = slot.iterator.next();
        if (done) {
            slot.iterator = undefined;
            return this.#getSlotClient(slotNumber);
        }
        return value;
    }
    #randomClientIterator;
    #getRandomClient() {
        if (!this.#clientByKey.size) {
            throw new Error('Cluster is not connected');
        }
        if (!this.#randomClientIterator) {
            this.#randomClientIterator = this.#clientByKey.values();
        }
        const { done, value } = this.#randomClientIterator.next();
        if (done) {
            this.#randomClientIterator = undefined;
            return this.#getRandomClient();
        }
        return value;
    }
    getClient(firstKey, isReadonly) {
        if (!firstKey) {
            return this.#getRandomClient();
        }
        const slot = cluster_key_slot_1.default(firstKey);
        if (!isReadonly || !this.#options.useReplicas) {
            return this.#getSlotMaster(slot);
        }
        return this.#getSlotClient(slot);
    }
    getMasters() {
        const masters = [];
        for (const client of this.#clientByKey.values()) {
            if (client.options?.readonly)
                continue;
            masters.push(client);
        }
        return masters;
    }
    async disconnect() {
        await Promise.all([...this.#clientByKey.values()].map(client => client.disconnect()));
        this.#clientByKey.clear();
        this.#slots.splice(0);
    }
}
exports.default = RedisClusterSlots;
