"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PubSubUnsubscribeCommands = exports.PubSubSubscribeCommands = void 0;
const yallist_1 = __importDefault(require("yallist"));
const redis_parser_1 = __importDefault(require("redis-parser"));
const errors_1 = require("./errors");
var PubSubSubscribeCommands;
(function (PubSubSubscribeCommands) {
    PubSubSubscribeCommands["SUBSCRIBE"] = "SUBSCRIBE";
    PubSubSubscribeCommands["PSUBSCRIBE"] = "PSUBSCRIBE";
})(PubSubSubscribeCommands = exports.PubSubSubscribeCommands || (exports.PubSubSubscribeCommands = {}));
var PubSubUnsubscribeCommands;
(function (PubSubUnsubscribeCommands) {
    PubSubUnsubscribeCommands["UNSUBSCRIBE"] = "UNSUBSCRIBE";
    PubSubUnsubscribeCommands["PUNSUBSCRIBE"] = "PUNSUBSCRIBE";
})(PubSubUnsubscribeCommands = exports.PubSubUnsubscribeCommands || (exports.PubSubUnsubscribeCommands = {}));
class RedisCommandsQueue {
    static encodeCommand(args) {
        const encoded = [
            `*${args.length}`,
            `$${args[0].length}`,
            args[0]
        ];
        for (let i = 1; i < args.length; i++) {
            const str = args[i].toString();
            encoded.push(`$${str.length}`, str);
        }
        return encoded.join('\r\n') + '\r\n';
    }
    static #flushQueue(queue, err) {
        while (queue.length) {
            queue.shift().reject(err);
        }
    }
    static #emitPubSubMessage(listeners, message, channel) {
        for (const listener of listeners) {
            listener(message, channel);
        }
    }
    #maxLength;
    #executor;
    #waitingToBeSent = new yallist_1.default();
    #waitingForReply = new yallist_1.default();
    #pubSubState = {
        subscribing: 0,
        subscribed: 0,
        unsubscribing: 0
    };
    #pubSubListeners = {
        channels: new Map(),
        patterns: new Map()
    };
    #parser = new redis_parser_1.default({
        returnReply: (reply) => {
            if ((this.#pubSubState.subscribing || this.#pubSubState.subscribed) && Array.isArray(reply)) {
                switch (reply[0]) {
                    case 'message':
                        return RedisCommandsQueue.#emitPubSubMessage(this.#pubSubListeners.channels.get(reply[1]), reply[2], reply[1]);
                    case 'pmessage':
                        return RedisCommandsQueue.#emitPubSubMessage(this.#pubSubListeners.patterns.get(reply[1]), reply[3], reply[2]);
                    case 'subscribe':
                    case 'psubscribe':
                        if (--this.#waitingForReply.head.value.channelsCounter === 0) {
                            this.#shiftWaitingForReply().resolve();
                        }
                        return;
                }
            }
            this.#shiftWaitingForReply().resolve(reply);
        },
        returnError: (err) => this.#shiftWaitingForReply().reject(err)
    });
    #chainInExecution;
    constructor(maxLength, executor) {
        this.#maxLength = maxLength;
        this.#executor = executor;
    }
    #isQueueBlocked() {
        if (this.#pubSubState.subscribing || this.#pubSubState.subscribed) {
            return Promise.reject(new Error('Cannot send commands in PubSub mode'));
        }
        else if (!this.#maxLength) {
            return;
        }
        return this.#waitingToBeSent.length + this.#waitingForReply.length >= this.#maxLength ?
            Promise.reject(new Error('The queue is full')) :
            undefined;
    }
    addCommand(args, options) {
        return this.#isQueueBlocked() || this.addEncodedCommand(RedisCommandsQueue.encodeCommand(args), options);
    }
    addEncodedCommand(encodedCommand, options) {
        const fullQueuePromise = this.#isQueueBlocked();
        if (fullQueuePromise) {
            return fullQueuePromise;
        }
        else if (options?.signal?.aborted) {
            return Promise.reject(new errors_1.AbortError());
        }
        return new Promise((resolve, reject) => {
            const node = new yallist_1.default.Node({
                encodedCommand,
                chainId: options?.chainId,
                resolve,
                reject
            });
            if (options?.signal) {
                const listener = () => {
                    this.#waitingToBeSent.removeNode(node);
                    node.value.reject(new errors_1.AbortError());
                };
                if (options.signal.aborted) {
                    return listener();
                }
                node.value.abort = {
                    signal: options.signal,
                    listener
                };
                options.signal.addEventListener('abort', listener, {
                    once: true
                });
            }
            if (options?.asap) {
                this.#waitingToBeSent.unshiftNode(node);
            }
            else {
                this.#waitingToBeSent.pushNode(node);
            }
        });
    }
    subscribe(command, channels, listener) {
        const channelsToSubscribe = [], listeners = command === PubSubSubscribeCommands.SUBSCRIBE ? this.#pubSubListeners.channels : this.#pubSubListeners.patterns;
        for (const channel of (Array.isArray(channels) ? channels : [channels])) {
            if (listeners.has(channel)) {
                listeners.get(channel).add(listener);
                continue;
            }
            listeners.set(channel, new Set([listener]));
            channelsToSubscribe.push(channel);
        }
        if (!channelsToSubscribe.length) {
            return Promise.resolve();
        }
        return this.#pushPubSubCommand(command, channelsToSubscribe);
    }
    unsubscribe(command, channels, listener) {
        const listeners = command === PubSubUnsubscribeCommands.UNSUBSCRIBE ? this.#pubSubListeners.channels : this.#pubSubListeners.patterns;
        if (!channels) {
            listeners.clear();
            return this.#pushPubSubCommand(command);
        }
        const channelsToUnsubscribe = [];
        for (const channel of (Array.isArray(channels) ? channels : [channels])) {
            const set = listeners.get(channel);
            if (!set)
                continue;
            let shouldUnsubscribe = !listener;
            if (listener) {
                set.delete(listener);
                shouldUnsubscribe = set.size === 0;
            }
            if (shouldUnsubscribe) {
                channelsToUnsubscribe.push(channel);
                listeners.delete(channel);
            }
        }
        if (!channelsToUnsubscribe.length) {
            return Promise.resolve();
        }
        return this.#pushPubSubCommand(command, channelsToUnsubscribe);
    }
    #pushPubSubCommand(command, channels) {
        return new Promise((resolve, reject) => {
            const isSubscribe = command === PubSubSubscribeCommands.SUBSCRIBE || command === PubSubSubscribeCommands.PSUBSCRIBE, inProgressKey = isSubscribe ? 'subscribing' : 'unsubscribing', commandArgs = [command];
            let channelsCounter;
            if (channels?.length) {
                commandArgs.push(...channels);
                channelsCounter = channels.length;
            }
            else {
                // unsubscribe only
                channelsCounter = (command[0] === 'P' ?
                    this.#pubSubListeners.patterns :
                    this.#pubSubListeners.channels).size;
            }
            this.#pubSubState[inProgressKey] += channelsCounter;
            this.#waitingToBeSent.push({
                encodedCommand: RedisCommandsQueue.encodeCommand(commandArgs),
                channelsCounter,
                resolve: () => {
                    this.#pubSubState[inProgressKey] -= channelsCounter;
                    this.#pubSubState.subscribed += channelsCounter * (isSubscribe ? 1 : -1);
                    resolve();
                },
                reject: () => {
                    this.#pubSubState[inProgressKey] -= channelsCounter;
                    reject();
                }
            });
        });
    }
    resubscribe() {
        if (!this.#pubSubState.subscribed && !this.#pubSubState.subscribing) {
            return;
        }
        this.#pubSubState.subscribed = this.#pubSubState.subscribing = 0;
        // TODO: acl error on one channel/pattern will reject the whole command
        return Promise.all([
            this.#pushPubSubCommand(PubSubSubscribeCommands.SUBSCRIBE, [...this.#pubSubListeners.channels.keys()]),
            this.#pushPubSubCommand(PubSubSubscribeCommands.PSUBSCRIBE, [...this.#pubSubListeners.patterns.keys()])
        ]);
    }
    executeChunk(recommendedSize) {
        if (!this.#waitingToBeSent.length)
            return;
        const encoded = [];
        let size = 0, lastCommandChainId;
        for (const command of this.#waitingToBeSent) {
            encoded.push(command.encodedCommand);
            size += command.encodedCommand.length;
            if (size > recommendedSize) {
                lastCommandChainId = command.chainId;
                break;
            }
        }
        if (!lastCommandChainId && encoded.length === this.#waitingToBeSent.length) {
            lastCommandChainId = this.#waitingToBeSent.tail.value.chainId;
        }
        lastCommandChainId ??= this.#waitingToBeSent.tail?.value.chainId;
        this.#executor(encoded.join(''));
        for (let i = 0; i < encoded.length; i++) {
            const waitingToBeSent = this.#waitingToBeSent.shift();
            if (waitingToBeSent.abort) {
                waitingToBeSent.abort.signal.removeEventListener('abort', waitingToBeSent.abort.listener);
            }
            this.#waitingForReply.push({
                resolve: waitingToBeSent.resolve,
                reject: waitingToBeSent.reject,
                channelsCounter: waitingToBeSent.channelsCounter
            });
        }
        this.#chainInExecution = lastCommandChainId;
    }
    parseResponse(data) {
        this.#parser.execute(data);
    }
    #shiftWaitingForReply() {
        if (!this.#waitingForReply.length) {
            throw new Error('Got an unexpected reply from Redis');
        }
        return this.#waitingForReply.shift();
    }
    flushWaitingForReply(err) {
        RedisCommandsQueue.#flushQueue(this.#waitingForReply, err);
        if (!this.#chainInExecution) {
            return;
        }
        while (this.#waitingToBeSent.head?.value.chainId === this.#chainInExecution) {
            this.#waitingToBeSent.shift();
        }
        this.#chainInExecution = undefined;
    }
    flushAll(err) {
        RedisCommandsQueue.#flushQueue(this.#waitingForReply, err);
        RedisCommandsQueue.#flushQueue(this.#waitingToBeSent, err);
    }
}
exports.default = RedisCommandsQueue;
;
