"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const commands_1 = __importDefault(require("./commands"));
const commands_queue_1 = __importDefault(require("./commands-queue"));
class RedisMultiCommand {
    static defineCommand(on, name, command) {
        on[name] = function (...args) {
            // do not return `this.addCommand` directly cause in legacy mode it's binded to the legacy version 
            this.addCommand(command.transformArguments(...args), command.transformReply);
            return this;
        };
    }
    static create(executor, clientOptions) {
        return new RedisMultiCommand(executor, clientOptions);
    }
    #executor;
    #clientOptions;
    #queue = [];
    #scriptsInUse = new Set();
    #modern = {};
    get modern() {
        if (!this.#clientOptions?.legacyMode) {
            throw new Error('client is not in "legacy mode"');
        }
        return this.#modern;
    }
    constructor(executor, clientOptions) {
        this.#executor = executor;
        this.#clientOptions = clientOptions;
        this.#initiateModules();
        this.#initiateScripts();
        this.#legacyMode();
    }
    #initiateModules() {
        if (!this.#clientOptions?.modules)
            return;
        for (const m of this.#clientOptions.modules) {
            for (const [name, command] of Object.entries(m)) {
                RedisMultiCommand.defineCommand(this, name, command);
            }
        }
    }
    #initiateScripts() {
        if (!this.#clientOptions?.scripts)
            return;
        for (const [name, script] of Object.entries(this.#clientOptions.scripts)) {
            this[name] = function (...args) {
                const evalArgs = [];
                if (this.#scriptsInUse.has(name)) {
                    evalArgs.push('EVALSHA', script.SHA);
                }
                else {
                    this.#scriptsInUse.add(name);
                    evalArgs.push('EVAL', script.SCRIPT);
                }
                return this.addCommand([
                    ...evalArgs,
                    script.NUMBER_OF_KEYS,
                    ...script.transformArguments(...args)
                ], script.transformReply);
            };
        }
    }
    #legacyMode() {
        if (!this.#clientOptions?.legacyMode)
            return;
        this.#modern.exec = this.exec.bind(this);
        this.#modern.addCommand = this.addCommand.bind(this);
        this.exec = function (...args) {
            const callback = typeof args[args.length - 1] === 'function' && args.pop();
            this.#modern.exec()
                .then((reply) => {
                if (!callback)
                    return;
                callback(null, reply);
            })
                .catch((err) => {
                if (!callback) {
                    // this.emit('error', err);
                    return;
                }
                callback(err);
            });
        };
        for (const name of Object.keys(commands_1.default)) {
            this.#defineLegacyCommand(name);
        }
        if (this.#clientOptions.modules) {
            for (const m of this.#clientOptions.modules) {
                for (const name of Object.keys(m)) {
                    this.#defineLegacyCommand(name);
                }
            }
        }
        if (this.#clientOptions.scripts) {
            for (const name of Object.keys(this.#clientOptions.scripts)) {
                this.#defineLegacyCommand(name);
            }
        }
    }
    #defineLegacyCommand(name) {
        this.#modern[name] = this[name];
        // TODO: https://github.com/NodeRedis/node-redis#commands:~:text=minimal%20parsing
        this[name] = function (...args) {
            return this.addCommand([name, ...args.flat()]);
        };
    }
    addCommand(args, transformReply) {
        this.#queue.push({
            encodedCommand: commands_queue_1.default.encodeCommand(args),
            transformReply
        });
        return this;
    }
    async exec(execAsPipeline = false) {
        if (execAsPipeline) {
            return this.execAsPipeline();
        }
        else if (!this.#queue.length) {
            return [];
        }
        const queue = this.#queue.splice(0);
        queue.unshift({
            encodedCommand: commands_queue_1.default.encodeCommand(['MULTI'])
        });
        queue.push({
            encodedCommand: commands_queue_1.default.encodeCommand(['EXEC'])
        });
        const rawReplies = await this.#executor(queue, Symbol('[RedisMultiCommand] Chain ID'));
        return rawReplies[rawReplies.length - 1].map((reply, i) => {
            const { transformReply } = queue[i + 1];
            return transformReply ? transformReply(reply) : reply;
        });
    }
    async execAsPipeline() {
        if (!this.#queue.length) {
            return [];
        }
        return await this.#executor(this.#queue.splice(0));
    }
}
exports.default = RedisMultiCommand;
for (const [name, command] of Object.entries(commands_1.default)) {
    RedisMultiCommand.defineCommand(RedisMultiCommand.prototype, name, command);
}
