"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = __importDefault(require("events"));
const net_1 = __importDefault(require("net"));
const tls_1 = __importDefault(require("tls"));
const promises_1 = require("timers/promises");
class RedisSocket extends events_1.default {
    static #initiateOptions(options) {
        options ??= {};
        if (!RedisSocket.#isUnixSocket(options)) {
            if (RedisSocket.#isUrlSocket(options)) {
                const url = new URL(options.url);
                options.port = Number(url.port);
                options.host = url.hostname;
                options.username = url.username;
                options.password = url.password;
            }
            options.port ??= 6379;
            options.host ??= '127.0.0.1';
        }
        return options;
    }
    static #defaultRetryStrategy(retries) {
        return Math.min(retries * 50, 500);
    }
    static #isUrlSocket(options) {
        return options.hasOwnProperty('url');
    }
    static #isUnixSocket(options) {
        return options.hasOwnProperty('path');
    }
    static #isTlsSocket(options) {
        return options.hasOwnProperty('tls');
    }
    #initiator;
    #options;
    #socket;
    #isOpen = false;
    get isOpen() {
        return this.#isOpen;
    }
    get chunkRecommendedSize() {
        if (!this.#socket)
            return 0;
        return this.#socket.writableHighWaterMark - this.#socket.writableLength;
    }
    constructor(initiator, options) {
        super();
        this.#initiator = initiator;
        this.#options = RedisSocket.#initiateOptions(options);
    }
    async connect() {
        if (this.#isOpen) {
            throw new Error('Socket is connection/connecting');
        }
        this.#isOpen = true;
        try {
            await this.#connect();
        }
        catch (err) {
            this.#isOpen = false;
            throw err;
        }
    }
    async #connect(hadError) {
        this.#socket = await this.#retryConnection(0, hadError);
        this.emit('connect');
        if (this.#initiator) {
            try {
                await this.#initiator();
            }
            catch (err) {
                this.#socket.end();
                this.#socket = undefined;
                throw err;
            }
        }
        this.emit('ready');
    }
    async #retryConnection(retries, hadError) {
        if (retries > 0 || hadError) {
            this.emit('reconnecting');
        }
        try {
            return await this.#createSocket();
        }
        catch (err) {
            this.emit('error', err);
            if (!this.#isOpen) {
                throw err;
            }
            const retryIn = (this.#options?.retryStrategy ?? RedisSocket.#defaultRetryStrategy)(retries);
            if (retryIn instanceof Error) {
                throw retryIn;
            }
            await promises_1.setTimeout(retryIn);
            return this.#retryConnection(retries + 1);
        }
    }
    #createSocket() {
        return new Promise((resolve, reject) => {
            const { connectEvent, socket } = RedisSocket.#isTlsSocket(this.#options) ?
                this.#createTlsSocket() :
                this.#createNetSocket();
            socket
                .once('error', (err) => reject(err))
                .once(connectEvent, () => {
                socket
                    .off('error', reject)
                    .once('error', (err) => this.#onSocketError(err))
                    .once('close', hadError => {
                    if (!hadError && this.#isOpen) {
                        this.#onSocketError(new Error('Socket closed unexpectedly'));
                    }
                })
                    .on('drain', () => this.emit('drain'))
                    .on('data', (data) => this.emit('data', data));
                resolve(socket);
            });
        });
    }
    #createNetSocket() {
        return {
            connectEvent: 'connect',
            socket: net_1.default.connect(this.#options) // TODO
        };
    }
    #createTlsSocket() {
        return {
            connectEvent: 'secureConnect',
            socket: tls_1.default.connect(this.#options) // TODO
        };
    }
    #onSocketError(err) {
        this.#socket = undefined;
        this.emit('error', err);
        this.#connect(true)
            .catch(err => this.emit('error', err));
    }
    write(encodedCommands) {
        if (!this.#socket) {
            throw new Error('Socket is closed');
        }
        return this.#socket.write(encodedCommands);
    }
    async disconnect() {
        if (!this.#isOpen || !this.#socket) {
            throw new Error('Socket is closed');
        }
        this.#isOpen = false;
        this.#socket.end();
        await events_1.default.once(this.#socket, 'end');
        this.#socket = undefined;
        this.emit('end');
    }
}
exports.default = RedisSocket;
