"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.waitTillBeenCalled = exports.itWithCluster = exports.itWithClient = exports.TEST_REDIS_CLUSTERES = exports.TestRedisClusters = exports.TEST_REDIS_SERVERS = exports.TestRedisServers = void 0;
const strict_1 = __importDefault(require("assert/strict"));
const client_1 = __importDefault(require("./client"));
const child_process_1 = require("child_process");
const events_1 = require("events");
const tcp_port_used_1 = __importDefault(require("tcp-port-used"));
const which_1 = __importDefault(require("which"));
const promises_1 = require("timers/promises");
const cluster_1 = __importDefault(require("./cluster"));
const promises_2 = require("fs/promises");
var TestRedisServers;
(function (TestRedisServers) {
    TestRedisServers[TestRedisServers["OPEN"] = 0] = "OPEN";
    TestRedisServers[TestRedisServers["PASSWORD"] = 1] = "PASSWORD";
})(TestRedisServers = exports.TestRedisServers || (exports.TestRedisServers = {}));
exports.TEST_REDIS_SERVERS = {};
var TestRedisClusters;
(function (TestRedisClusters) {
    TestRedisClusters[TestRedisClusters["OPEN"] = 0] = "OPEN";
})(TestRedisClusters = exports.TestRedisClusters || (exports.TestRedisClusters = {}));
exports.TEST_REDIS_CLUSTERES = {};
before(function () {
    this.timeout(10000);
    return Promise.all([
        spawnOpenServer(),
        spawnPasswordServer(),
        spawnOpenCluster()
    ]);
});
async function spawnOpenServer() {
    exports.TEST_REDIS_SERVERS[TestRedisServers.OPEN] = {
        port: await spawnRedisServer()
    };
}
async function spawnPasswordServer() {
    exports.TEST_REDIS_SERVERS[TestRedisServers.PASSWORD] = {
        port: await spawnRedisServer(['--requirepass', 'password']),
        username: 'default',
        password: 'password'
    };
}
async function spawnOpenCluster() {
    exports.TEST_REDIS_CLUSTERES[TestRedisClusters.OPEN] = (await spawnRedisCluster(TestRedisClusters.OPEN, 3)).map(port => ({
        port
    }));
}
function itWithClient(type, title, fn) {
    it(title, async () => {
        const client = client_1.default.create({
            socket: exports.TEST_REDIS_SERVERS[type]
        });
        await client.connect();
        try {
            await client.flushAll();
            await fn(client);
        }
        finally {
            await client.flushAll();
            await client.disconnect();
        }
    });
}
exports.itWithClient = itWithClient;
function itWithCluster(type, title, fn) {
    it(title, async () => {
        const cluster = cluster_1.default.create({
            rootNodes: exports.TEST_REDIS_CLUSTERES[type]
        });
        await cluster.connect();
        try {
            await clusterFlushAll(cluster);
            await fn(cluster);
        }
        finally {
            await clusterFlushAll(cluster);
            await cluster.disconnect();
        }
    });
}
exports.itWithCluster = itWithCluster;
async function clusterFlushAll(cluster) {
    await Promise.all(cluster.getMasters().map(master => master.flushAll()));
}
const REDIS_PATH = which_1.default.sync('redis-server');
let port = 6379;
async function spawnRedisServer(args) {
    const currentPort = port++, process = child_process_1.spawn(REDIS_PATH, [
        '--save',
        '',
        '--port',
        currentPort.toString(),
        ...(args ?? [])
    ]);
    // TODO: catch process exit
    await tcp_port_used_1.default.waitForStatus(currentPort, '127.0.0.1', true, 10, 10000);
    after(() => {
        strict_1.default.ok(process.kill());
        return events_1.once(process, 'close');
    });
    return currentPort;
}
const SLOTS = 16384, CLUSTER_NODE_TIMEOUT = 2000;
async function spawnRedisCluster(type, numberOfNodes, args) {
    const spawnPromises = [], slotsPerNode = Math.floor(SLOTS / numberOfNodes);
    for (let i = 0; i < numberOfNodes; i++) {
        const fromSlot = i * slotsPerNode;
        spawnPromises.push(spawnRedisClusterNode(type, i, fromSlot, i === numberOfNodes - 1 ? SLOTS : fromSlot + slotsPerNode, args));
    }
    const spawnResults = await Promise.all(spawnPromises), meetPromises = [];
    for (let i = 1; i < spawnResults.length; i++) {
        meetPromises.push(spawnResults[i].client.clusterMeet('127.0.0.1', spawnResults[i - 1].port));
    }
    while ((await spawnResults[0].client.clusterInfo()).state !== 'ok') {
        await promises_1.setTimeout(CLUSTER_NODE_TIMEOUT);
    }
    const ports = [], disconnectPromises = [];
    for (const { port, client } of spawnResults) {
        ports.push(port);
        disconnectPromises.push(client.disconnect());
    }
    await Promise.all(disconnectPromises);
    return ports;
}
async function spawnRedisClusterNode(type, nodeIndex, fromSlot, toSlot, args) {
    const clusterConfigFile = `/tmp/${type}-${nodeIndex}.conf`, port = await spawnRedisServer([
        '--cluster-enabled',
        'yes',
        '--cluster-node-timeout',
        CLUSTER_NODE_TIMEOUT.toString(),
        '--cluster-config-file',
        clusterConfigFile,
        ...(args ?? [])
    ]);
    after(async () => {
        try {
            await promises_2.unlink(clusterConfigFile);
        }
        catch (err) {
            if (err.code == 'ENOENT')
                return;
            throw err;
        }
    });
    const client = client_1.default.create({
        socket: {
            port
        }
    });
    await client.connect();
    const range = [];
    for (let i = fromSlot; i < toSlot; i++) {
        range.push(i);
    }
    await Promise.all([
        client.clusterFlushSlots(),
        client.clusterAddSlots(range)
    ]);
    return {
        port,
        client
    };
}
async function waitTillBeenCalled(spy) {
    const start = process.hrtime.bigint(), calls = spy.callCount;
    do {
        if (process.hrtime.bigint() - start > 1_000_000_000) {
            throw new Error('Waiting for more than 1 second');
        }
        await promises_1.setTimeout(1);
    } while (spy.callCount === calls);
}
exports.waitTillBeenCalled = waitTillBeenCalled;
